//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "App_Open.h"
#include "App_Processes.h"
#include "App_ProgressDialog.h"
#include "Moa_Tasks.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace App_open {
//---------------------------------------------------------------------------
typedef void __fastcall (__closure * TOpenNextEvent)(TObject* ASender, TProgramEntry* AProgramEntry, TDocumentEntry* ADocumentEntry);
typedef void __fastcall (__closure * TOpenMessageEvent)(TObject* ASender, const AnsiString& AMessage);
//---------------------------------------------------------------------------
enum TOpenStep
{
	osNone,
	osPrepare,
	osOpen,
	osOpenFail,
	osWait,
	osWaitFail,
	osNoWait,
	osUnprepare,
};
//---------------------------------------------------------------------------
class TOpenTask : public TTask
{
private:
	class EOpenSkip : public EAbort
	{
	public:
		__fastcall EOpenSkip(const AnsiString& AMessage) : EAbort(AMessage) {}
	};
	struct TPrepareArguments
	{
		bool Skip;
	};
	struct TMarkArguments
	{
		TOpenResult Result;
	};
	struct TOnNextArguments
	{
		TProgramEntry* ProgramEntry;
		TDocumentEntry* DocumentEntry;
	};
	struct TOnInformationArguments
	{
		AnsiString Message;
	};
	typedef TOnInformationArguments TOnWarningArguments;
	typedef TOnInformationArguments TOnErrorArguments;
	TProgramEntry* FProgramEntry;
	TDocumentEntriesModel* FDocumentEntriesModel;
	TProcessCreator* FProcessCreator;
	TOpenStep FStep;
	int FCurrent;
	int FAll;
	bool FContinueRequested;
	TOpenNextEvent FOnNext;
	TNotifyEvent FOnPrepare;
	TNotifyEvent FOnOpen;
	TNotifyEvent FOnOpenFail;
	TNotifyEvent FOnWait;
	TNotifyEvent FOnWaitFail;
	TNotifyEvent FOnNoWait;
	TNotifyEvent FOnUnprepare;
	TOpenMessageEvent FOnInformation;
	TOpenMessageEvent FOnWarning;
	TOpenMessageEvent FOnError;
	__property TProgramEntry* ProgramEntry = {read=FProgramEntry};
	__property TDocumentEntriesModel* DocumentEntriesModel = {read=FDocumentEntriesModel};
	__property TProcessCreator* ProcessCreator = {read=FProcessCreator};
	__property TOpenStep Step = {read=FStep, write=SetStep};
	__property int Current = {read=FCurrent};
	__property int All = {read=FAll};
	__property bool ContinueRequested = {read=FContinueRequested};
	void __fastcall Initialize();
	void __fastcall Uninitialize();
	void __fastcall Process();
	void __fastcall Prepare();
	void __fastcall Unprepare();
	void __fastcall Mark(TOpenResult AResult);
	void __fastcall SynchronizedPrepare(void* AData);
	void __fastcall SynchronizedUnprepare(void* AData);
	void __fastcall SynchronizedMark(void* AData);
	void __fastcall SynchronizedDoNext(void* AData);
	void __fastcall SynchronizedDoPrepare(void* AData);
	void __fastcall SynchronizedDoOpen(void* AData);
	void __fastcall SynchronizedDoOpenFail(void* AData);
	void __fastcall SynchronizedDoWait(void* AData);
	void __fastcall SynchronizedDoWaitFail(void* AData);
	void __fastcall SynchronizedDoNoWait(void* AData);
	void __fastcall SynchronizedDoUnprepare(void* AData);
	void __fastcall SynchronizedDoInformation(void* AData);
	void __fastcall SynchronizedDoWarning(void* AData);
	void __fastcall SynchronizedDoError(void* AData);
	void __fastcall SetStep(TOpenStep AStep);
protected:
	virtual void __fastcall Execute();
	virtual void __fastcall DoNext(TProgramEntry* AProgramEntry, TDocumentEntry* ADocumentEntry);
	virtual void __fastcall DoPrepare();
	virtual void __fastcall DoOpen();
	virtual void __fastcall DoOpenFail();
	virtual void __fastcall DoWait();
	virtual void __fastcall DoWaitFail();
	virtual void __fastcall DoNoWait();
	virtual void __fastcall DoUnprepare();
	virtual void __fastcall DoInformation(const AnsiString& AMessage);
	virtual void __fastcall DoWarning(const AnsiString& AMessage);
	virtual void __fastcall DoError(const AnsiString& AMessage);
public:
	__fastcall TOpenTask(TProgramEntry* AProgramEntry, TDocumentEntriesModel* ADocumentEntriesModel);
	virtual __fastcall ~TOpenTask();
	virtual void __fastcall Continue();
	__property TOpenNextEvent OnNext = {read=FOnNext, write=FOnNext};
	__property TNotifyEvent OnPrepare = {read=FOnPrepare, write=FOnPrepare};
	__property TNotifyEvent OnOpen = {read=FOnOpen, write=FOnOpen};
	__property TNotifyEvent OnOpenFail = {read=FOnOpenFail, write=FOnOpenFail};
	__property TNotifyEvent OnWait = {read=FOnWait, write=FOnWait};
	__property TNotifyEvent OnWaitFail = {read=FOnWaitFail, write=FOnWaitFail};
	__property TNotifyEvent OnNoWait = {read=FOnNoWait, write=FOnNoWait};
	__property TNotifyEvent OnUnprepare = {read=FOnUnprepare, write=FOnUnprepare};
	__property TOpenMessageEvent OnInformation = {read=FOnInformation, write=FOnInformation};
	__property TOpenMessageEvent OnWarning = {read=FOnWarning, write=FOnWarning};
	__property TOpenMessageEvent OnError = {read=FOnError, write=FOnError};
};
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "App_Processes.h"
#include "App_ProgressDialog.h"
#include "Moa_AutoPointers.h"
#include "Moa_DynamicArrays.h"
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Initialize()
{
	try {
		FProcessCreator = new TProcessCreator();
	}
	catch(...) {
		Uninitialize();
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Uninitialize()
{
	delete FProcessCreator;
	FProcessCreator = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Process()
{
	try {
		Prepare();
		Step = osOpen;
		TAutoPointer<TProcess> LProcess = ProcessCreator->Execute();
		Mark(orSuccess);
		Step = osWait;
		while(!Canceled && !LProcess->IsTerminated(100)) {}
	}
	catch(EOpenSkip& LException) {
		// Do nothing
	}
	catch(EProcessCreateError& LException) {
		Mark(orFailure);
		Step = osOpenFail;
		while(!Canceled && !ContinueRequested) {::Sleep(100);}
	}
	catch(EProcessUnavailable& LException) {
		Mark(orSuccess);
		Step = osNoWait;
		while(!Canceled && !ContinueRequested) {::Sleep(100);}
	}
	catch(EProcessWaitFail& LException) {
		Step = osWaitFail;
		while(!Canceled && !ContinueRequested) {::Sleep(100);}
	}
	Unprepare();
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Prepare()
{
	TPrepareArguments LArguments = {false};
	Synchronize(&SynchronizedPrepare, &LArguments);
	if(LArguments.Skip) {
		throw EOpenSkip("The program skips the document.");
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Unprepare()
{
	Synchronize(&SynchronizedUnprepare, NULL);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Mark(TOpenResult AResult)
{
	TMarkArguments LArguments = {AResult};
	Synchronize(&SynchronizedMark, &LArguments);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedPrepare(void* AData)
{
	TDocumentEntry* LDocumentEntry = DocumentEntriesModel->Items[Current];
	DoNext(ProgramEntry, LDocumentEntry);
	Step = osPrepare;
	{
		TPrepareArguments& LArguments = *static_cast<TPrepareArguments*>(AData);
		LArguments.Skip = LDocumentEntry->Result == orSuccess;
		if(LArguments.Skip) {
			return;
		}
	}
	if(ProgramEntry->IsDefault()) {
		ProcessCreator->File = LDocumentEntry->Path;
		ProcessCreator->Parameters = AnsiString();
		ProcessCreator->WorkingDirectory = ExtractFileDir(LDocumentEntry->Path);
		ProcessCreator->ShowCommand = SW_SHOWNORMAL;
		ProcessCreator->ReturnProcess = true;
	}
	else {
		ProcessCreator->File = ProgramEntry->File;
		ProcessCreator->Parameters = ProgramEntry->DocumentPrefix + LDocumentEntry->Path + ProgramEntry->DocumentSuffix;
		ProcessCreator->WorkingDirectory = ProgramEntry->WorkingDirectory;
		ProcessCreator->ShowCommand = ProgramEntry->ShowCommand;
		ProcessCreator->ReturnProcess = ProgramEntry->WaitForTermination;
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedUnprepare(void* AData)
{
	Step = osUnprepare;
	++FCurrent;
	DoProgress(Current, All);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedMark(void* AData)
{
	TMarkArguments& LArguments = *static_cast<TMarkArguments*>(AData);
	DocumentEntriesModel->Items[Current]->Result = LArguments.Result;
	DocumentEntriesModel->ItemsChange(TDynamicArrayUtilities<int>::FromElements(Current));
}
//---------------------------------------------------------------------------

void __fastcall TOpenTask::SynchronizedDoNext(void* AData)
{
	TOnNextArguments& LArguments = *static_cast<TOnNextArguments*>(AData);
	OnNext(this, LArguments.ProgramEntry, LArguments.DocumentEntry);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoPrepare(void* AData)
{
	OnPrepare(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoOpen(void* AData)
{
	OnOpen(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoOpenFail(void* AData)
{
	OnOpenFail(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoWait(void* AData)
{
	OnWait(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoWaitFail(void* AData)
{
	OnWaitFail(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoNoWait(void* AData)
{
	OnNoWait(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoUnprepare(void* AData)
{
	OnUnprepare(this);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoInformation(void* AData)
{
	TOnInformationArguments& LArguments = *static_cast<TOnInformationArguments*>(AData);
	OnInformation(this, LArguments.Message);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoWarning(void* AData)
{
	TOnWarningArguments& LArguments = *static_cast<TOnWarningArguments*>(AData);
	OnWarning(this, LArguments.Message);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SynchronizedDoError(void* AData)
{
	TOnErrorArguments& LArguments = *static_cast<TOnErrorArguments*>(AData);
	OnError(this, LArguments.Message);
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::SetStep(TOpenStep AStep)
{
	if(AStep == Step) {
		return;
	}
	FStep = AStep;
	switch(Step) {
	case osNone:
		// Do nothing
		break;
	case osPrepare:
		DoPrepare();
		DoInformation("Preparing...");
		break;
	case osOpen:
		DoOpen();
		DoInformation("Opening a document...");
		break;
	case osOpenFail:
		DoOpenFail();
		DoError("Could not open the document");
		break;
	case osWait:
		DoWait();
		DoInformation("Waiting for the program to quit...");
		break;
	case osWaitFail:
		DoWaitFail();
		DoError("Could not wait for the program to quit");
		break;
	case osNoWait:
		DoNoWait();
		DoWarning("Cannot wait for the program to quit");
		break;
	case osUnprepare:
		DoUnprepare();
		DoInformation("Unpreparing...");
		break;
	default:
		throw EAppError("The program does not support the step.");
	}
	FContinueRequested = false;
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Execute()
{
	while(!Canceled && Current < All) {
		Process();
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoNext(TProgramEntry* AProgramEntry, TDocumentEntry* ADocumentEntry)
{
	if(OnNext != NULL) {
		TOnNextArguments LArguments = {AProgramEntry, ADocumentEntry};
		Synchronize(&SynchronizedDoNext, &LArguments);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoPrepare()
{
	if(OnPrepare != NULL) {
		Synchronize(&SynchronizedDoPrepare, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoOpen()
{
	if(OnOpen != NULL) {
		Synchronize(&SynchronizedDoOpen, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoOpenFail()
{
	if(OnOpenFail != NULL) {
		Synchronize(&SynchronizedDoOpenFail, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoWait()
{
	if(OnWait != NULL) {
		Synchronize(&SynchronizedDoWait, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoWaitFail()
{
	if(OnWaitFail != NULL) {
		Synchronize(&SynchronizedDoWaitFail, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoNoWait()
{
	if(OnNoWait != NULL) {
		Synchronize(&SynchronizedDoNoWait, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoUnprepare()
{
	if(OnUnprepare != NULL) {
		Synchronize(&SynchronizedDoUnprepare, NULL);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoInformation(const AnsiString& AMessage)
{
	if(OnInformation != NULL) {
		TOnInformationArguments LArguments;
		LArguments.Message = AMessage;
		Synchronize(&SynchronizedDoInformation, &LArguments);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoWarning(const AnsiString& AMessage)
{
	if(OnWarning != NULL) {
		TOnWarningArguments LArguments;
		LArguments.Message = AMessage;
		Synchronize(&SynchronizedDoWarning, &LArguments);
	}
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::DoError(const AnsiString& AMessage)
{
	if(OnError != NULL) {
		TOnErrorArguments LArguments;
		LArguments.Message = AMessage;
		Synchronize(&SynchronizedDoError, &LArguments);
	}
}
//---------------------------------------------------------------------------
__fastcall TOpenTask::TOpenTask(TProgramEntry* AProgramEntry, TDocumentEntriesModel* ADocumentEntriesModel) :
	TTask(),
	FProgramEntry(AProgramEntry),
	FDocumentEntriesModel(ADocumentEntriesModel),
	FProcessCreator(NULL),
	FStep(osNone),
	FCurrent(0),
	FAll(ADocumentEntriesModel->Count),
	FContinueRequested(false),
	FOnNext(NULL),
	FOnPrepare(NULL),
	FOnOpen(NULL),
	FOnOpenFail(NULL),
	FOnWait(NULL),
	FOnWaitFail(NULL),
	FOnNoWait(NULL),
	FOnUnprepare(NULL),
	FOnInformation(NULL),
	FOnWarning(NULL),
	FOnError(NULL)
{
	Initialize();
}
//---------------------------------------------------------------------------
__fastcall TOpenTask::~TOpenTask()
{
	Uninitialize();
}
//---------------------------------------------------------------------------
void __fastcall TOpenTask::Continue()
{
	FContinueRequested = true;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskTerminate(TObject* ASender)
{
	ProgressDialog->Finish();
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskProgress(TObject* ASender, int ACurrent, int AAll)
{
	ProgressDialog->ShowProgress(ACurrent, AAll);
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskNext(TObject* ASender, TProgramEntry* AProgramEntry, TDocumentEntry* ADocumentEntry)
{
	ProgressDialog->ShowParameters(AProgramEntry->DisplayName, ExtractFileName(ADocumentEntry->Path));
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskPrepare(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskOpen(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskOpenFail(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskWait(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskWaitFail(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskNoWait(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = true;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskUnprepare(TObject* ASender)
{
	ProgressDialog->ContinueButton->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskInformation(TObject* ASender, const AnsiString& AMessage)
{
	ProgressDialog->ShowInformation(AMessage);
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskWarning(TObject* ASender, const AnsiString& AMessage)
{
	ProgressDialog->ShowWarning(AMessage);
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::TaskError(TObject* ASender, const AnsiString& AMessage)
{
	ProgressDialog->ShowError(AMessage);
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::ProgressDialogContinue(TObject* ASender)
{
	Task->Continue();
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::ProgressDialogCancel(TObject* ASender)
{
	Task->Cancel();
}
//---------------------------------------------------------------------------
__fastcall TOpenCommands::TOpenCommands(
	TProgramEntriesModel* AProgramEntriesModel, 
	TProgramEntriesSelection* AProgramEntriesSelection, 
	TDocumentEntriesModel* ADocumentEntriesModel) :
	TCommands(),
	FProgramEntriesModel(AProgramEntriesModel),
	FProgramEntriesSelection(AProgramEntriesSelection),
	FDocumentEntriesModel(ADocumentEntriesModel),
	FTask(NULL)
{
	// Do nothing
}
//---------------------------------------------------------------------------
__fastcall TOpenCommands::~TOpenCommands()
{
	// Do nothing
}
//---------------------------------------------------------------------------
void __fastcall TOpenCommands::Start()
{
	if(!CanStart()) {
		HandleFeatureUnavailable();
	}
	try {
		TAutoPointer<TOpenTask> LTask = new TOpenTask(
			ProgramEntriesModel->Items[ProgramEntriesSelection->SelectionIndex],
			DocumentEntriesModel);
		LTask->FreeOnTerminate = true;
		LTask->OnTerminate = TaskTerminate;
		LTask->OnProgress = TaskProgress;
		LTask->OnNext = TaskNext;
		LTask->OnPrepare = TaskPrepare;
		LTask->OnOpen = TaskOpen;
		LTask->OnOpenFail = TaskOpenFail;
		LTask->OnWait = TaskWait;
		LTask->OnWaitFail = TaskWaitFail;
		LTask->OnNoWait = TaskNoWait;
		LTask->OnUnprepare = TaskUnprepare;
		LTask->OnInformation = TaskInformation;
		LTask->OnWarning = TaskWarning;
		LTask->OnError = TaskError;
		LTask->Perform();
		FTask = LTask.Release();
		TProgressDialog::EnsureCreated();
		ProgressDialog->Caption = "Open";
		ProgressDialog->Reset();
		ProgressDialog->OnContinue = ProgressDialogContinue;
		ProgressDialog->OnCancel = ProgressDialogCancel;
		ProgressDialog->Execute();
	}
	__finally {
		FTask = NULL;
	}
}
//---------------------------------------------------------------------------
bool __fastcall TOpenCommands::CanStart()
{
	return
		ProgramEntriesSelection->SelectionIndex >= 0 &&
		DocumentEntriesModel->Count > 0;
}
//---------------------------------------------------------------------------
} // End namespace App_open
//---------------------------------------------------------------------------

