//---------------------------------------------------------------------------
// Version: 1.2.0
// Date: 2012-06-02 UTC+09:00
// Author: Moochos
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "Moa_AutoPointers.h"
#include "Moa_ShellFiles.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace Moa_shellfiles {
//---------------------------------------------------------------------------
UINT __fastcall TShellFileUtilities::IconScaleToFlags(TIconScale AIconScale)
{
	switch(AIconScale) {
	case isShellSized:
		return SHGFI_SHELLICONSIZE;
	case isSmall:
		return SHGFI_SMALLICON;
	case isLarge:
		return SHGFI_LARGEICON;
	default:
		throw EShellFileError("The program does not support the icon scale.");
	}
}
//---------------------------------------------------------------------------
UINT __fastcall TShellFileUtilities::IconEffectsToFlags(const TIconEffects& AIconEffects)
{
	UINT LFlags = 0;
	if(AIconEffects.Contains(ieLinkOverlay)) {
		LFlags |= SHGFI_LINKOVERLAY;
	}
	if(AIconEffects.Contains(ieSelected)) {
		LFlags |= SHGFI_SELECTED;
	}
	return LFlags;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TShellFileUtilities::EnsureValidExtension(const AnsiString& AExtension)
{
	if(AExtension.Length() > 0 && AExtension[1] != '.') {
		return AnsiString('.') + AExtension;
	}
	return AExtension;
}
//---------------------------------------------------------------------------
const TIconScale TShellFileUtilities::DefaultIconScale = isShellSized;
//---------------------------------------------------------------------------
HIMAGELIST __fastcall TShellFileUtilities::GetImageList()
{
	return GetImageList(DefaultIconScale);
}
//---------------------------------------------------------------------------
HIMAGELIST __fastcall TShellFileUtilities::GetImageList(TIconScale AIconScale)
{
	SHFILEINFO LFileInfo = {0};
	HIMAGELIST LImageList = reinterpret_cast<HIMAGELIST>(::SHGetFileInfo(
		"",
		0,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | SHGFI_SYSICONINDEX));
	if(LImageList == NULL) {
		throw EShellFileError("The program could not get the image list.");
	}
	return LImageList;
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIcon(int AIndex, TIcon* AIcon)
{
	return GetIcon(AIndex, AIcon, DefaultIconScale);
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIcon(int AIndex, TIcon* AIcon, TIconScale AIconScale)
{
	return GetIcon(AIndex, AIcon, AIconScale, ILD_TRANSPARENT);
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIcon(int AIndex, TIcon* AIcon, TIconScale AIconScale, UINT ADrawStyle)
{
	HICON LIcon = ::ImageList_GetIcon(
		GetImageList(AIconScale),
		AIndex,
		ADrawStyle);
	if(LIcon == NULL) {
		throw EShellFileError("The program could not get the icon.");
	}
	try {
		AIcon->Handle = LIcon;
	}
	catch(...) {
		::DestroyIcon(LIcon);
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByPath(const AnsiString& APath, TIcon* AIcon)
{
	return GetIconByPath(APath, AIcon, DefaultIconScale);
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByPath(const AnsiString& APath, TIcon* AIcon, TIconScale AIconScale)
{
	return GetIconByPath(APath, AIcon, DefaultIconScale, TIconEffects());
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByPath(const AnsiString& APath, TIcon* AIcon, TIconScale AIconScale, const TIconEffects& AIconEffects)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		APath.c_str(),
		0,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | SHGFI_ICON) == 0) {
		throw EShellFileError("The program could not get the icon.");
	}
	try {
		AIcon->Handle = LFileInfo.hIcon;
	}
	catch(...) {
		::DestroyIcon(LFileInfo.hIcon);
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByExtension(const AnsiString& AExtension, TIcon* AIcon)
{
	GetIconByExtension(AExtension, AIcon, DefaultIconScale);
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByExtension(const AnsiString& AExtension, TIcon* AIcon, TIconScale AIconScale)
{
	GetIconByExtension(AExtension, AIcon, AIconScale, TIconEffects());
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconByExtension(const AnsiString& AExtension, TIcon* AIcon, TIconScale AIconScale, const TIconEffects& AIconEffects)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		("BaseName" + EnsureValidExtension(AExtension)).c_str(),
		FILE_ATTRIBUTE_NORMAL,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | IconEffectsToFlags(AIconEffects) | SHGFI_ICON | SHGFI_USEFILEATTRIBUTES) == 0) {
		throw EShellFileError("The program could not get the icon.");
	}
	try {
		AIcon->Handle = LFileInfo.hIcon;
	}
	catch(...) {
		::DestroyIcon(LFileInfo.hIcon);
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconForDirectory(TIcon* AIcon)
{
	GetIconForDirectory(AIcon, DefaultIconScale);
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconForDirectory(TIcon* AIcon, TIconScale AIconScale)
{
	GetIconForDirectory(AIcon, AIconScale, TIconEffects());
}
//---------------------------------------------------------------------------
void __fastcall TShellFileUtilities::GetIconForDirectory(TIcon* AIcon, TIconScale AIconScale, const TIconEffects& AIconEffects)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		"Name",
		FILE_ATTRIBUTE_DIRECTORY,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | IconEffectsToFlags(AIconEffects) | SHGFI_ICON | SHGFI_USEFILEATTRIBUTES) == 0) {
		throw EShellFileError("The program could not get the icon.");
	}
	try {
		AIcon->Handle = LFileInfo.hIcon;
	}
	catch(...) {
		::DestroyIcon(LFileInfo.hIcon);
		throw;
	}
}
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconByPath(const AnsiString& APath)
{
	return FindIconByPath(APath, DefaultIconScale);
}
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconByPath(const AnsiString& APath, TIconScale AIconScale)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		APath.c_str(),
		0,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | SHGFI_SYSICONINDEX) == 0) {
		throw EShellFileError("The program could not find the icon.");
	}
	return LFileInfo.iIcon;
}                                                   
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconByExtension(const AnsiString& AExtension)
{
	return FindIconByExtension(AExtension, DefaultIconScale);
}
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconByExtension(const AnsiString& AExtension, TIconScale AIconScale)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		("BaseName" + EnsureValidExtension(AExtension)).c_str(),
		FILE_ATTRIBUTE_NORMAL,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | SHGFI_SYSICONINDEX | SHGFI_USEFILEATTRIBUTES) == 0) {
		throw EShellFileError("The program could not find the icon.");
	}
	return LFileInfo.iIcon;
}
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconForDirectory()
{
	return FindIconForDirectory(DefaultIconScale);
}
//---------------------------------------------------------------------------
int __fastcall TShellFileUtilities::FindIconForDirectory(TIconScale AIconScale)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		"Name",
		FILE_ATTRIBUTE_DIRECTORY,
		&LFileInfo,
		sizeof(SHFILEINFO),
		IconScaleToFlags(AIconScale) | SHGFI_SYSICONINDEX | SHGFI_USEFILEATTRIBUTES) == 0) {
		throw EShellFileError("The program could not find the icon.");
	}
	return LFileInfo.iIcon;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
AnsiString __fastcall TFileTypeCache::TEntry::Determine(const AnsiString& AFile, bool AAsDirectory)
{
	SHFILEINFO LFileInfo = {0};
	if(::SHGetFileInfo(
		AFile.c_str(),
		AAsDirectory ? FILE_ATTRIBUTE_DIRECTORY : FILE_ATTRIBUTE_NORMAL,
		&LFileInfo,
		sizeof(SHFILEINFO),
		SHGFI_TYPENAME | SHGFI_USEFILEATTRIBUTES) == 0) {
		throw EShellFileError("The program could not get the file type.");
	}
	return LFileInfo.szTypeName;
}
//---------------------------------------------------------------------------
__fastcall TFileTypeCache::TEntry::TEntry() :
	TObject(),
	FDescription(Determine(GetCurrentDir(), true))
{
	// Do nothing
}
//---------------------------------------------------------------------------
__fastcall TFileTypeCache::TEntry::TEntry(const AnsiString& AFile) :
	TObject(),
	FDescription(Determine(AFile, false))
{
	// Do nothing
}
//---------------------------------------------------------------------------
__fastcall TFileTypeCache::TEntry::~TEntry()
{
	// Do nothing
}
//---------------------------------------------------------------------------
void __fastcall TFileTypeCache::Initialize()
{
	try {
		FExtensions = new TStringList();
		FExtensions->Sorted = true;
	}
	catch(...) {
		Uninitialize();
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TFileTypeCache::Uninitialize()
{
	delete FDirectory;
	FDirectory = NULL;

	if(FExtensions != NULL) {
		for(int LI = 0; LI < FExtensions->Count; ++LI) {
			delete FExtensions->Objects[LI];
		}
		delete FExtensions;
		FExtensions = NULL;
	}
}
//---------------------------------------------------------------------------
__fastcall TFileTypeCache::TFileTypeCache() :
	TObject(),
	FExtensions(NULL),
	FDirectory(NULL)
{
	Initialize();
}
//---------------------------------------------------------------------------
__fastcall TFileTypeCache::~TFileTypeCache()
{
	Uninitialize();
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFileTypeCache::Determine(const AnsiString& AFile)
{
	DWORD LFileAttributes = ::GetFileAttributes(AFile.c_str());
	if(LFileAttributes == 0xFFFFFFFF) {
		throw EShellFileError("The program cannot determine the file type.");
	}
	if((LFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
		return DetermineAsDirectory();
	}
	else {
		return DetermineAsRegularFile(AFile);
	}
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFileTypeCache::DetermineAsRegularFile(const AnsiString& AFile)
{
	AnsiString LFile = ExpandFileName(AFile);
	AnsiString LExtension = AnsiUpperCaseFileName(ExtractFileExt(LFile));
	int LIndex;
	if(Extensions->Find(LExtension, LIndex)) {
		return static_cast<TEntry*>(Extensions->Objects[LIndex])->Description;
	}
	TAutoPointer<TEntry> LEntry = new TEntry(LFile);
	Extensions->AddObject(LExtension, LEntry);
	return LEntry.Release()->Description;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFileTypeCache::DetermineAsDirectory()
{
	if(Directory == NULL) {
		FDirectory = new TEntry();
	}
	return Directory->Description;
}
//---------------------------------------------------------------------------
void __fastcall TFileTypeCache::Clear()
{
	delete FDirectory;
	FDirectory = NULL;

	for(int LI = 0; LI < Extensions->Count; ++LI) {
		delete Extensions->Objects[LI];
	}
	Extensions->Clear();
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::Determine(UINT AId)
{
	TAutoArray<TCHAR> LBuffer = new TCHAR[MAX_PATH];
	::ZeroMemory(LBuffer, sizeof(TCHAR) * MAX_PATH);
	if(!SHGetSpecialFolderPath(NULL, LBuffer, AId, false)) {
		throw EShellFileError("The program could not get the path of the special file.");
	}
	return AnsiString(LBuffer);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetApplicationData() 
{
	return Determine(CSIDL_APPDATA);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetDesktop() 
{
	return Determine(CSIDL_DESKTOPDIRECTORY);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetDocuments() 
{
	return Determine(CSIDL_PERSONAL);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetFavorites() 
{
	return Determine(CSIDL_FAVORITES);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetProgramFiles() 
{
	return Determine(CSIDL_PROGRAM_FILES);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetPrograms() 
{
	return Determine(CSIDL_PROGRAMS);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSpecialShellFiles::GetStartMenu()
{
	return Determine(CSIDL_STARTMENU);
}
//---------------------------------------------------------------------------
} // namespace Moa_shellfiles
//---------------------------------------------------------------------------

